
#define DEBUG
/*------------------ CIRCUITO ----------------------------------------------------------------------------------------------------------------------------------------------
 * PIN 2 - Entrada DCC desde opto
 * PIN A2 - LED de control (R470 + LED con cátodo a masa)
 * PIN A0 - Pulsador (R10K a +5V y Pulsador a masa)
 * Se montan los LED a masa, los LED se montan en ánodo común a +5V
 * PIN LED - (R470 + LED ánodo a +5V en cada salida)
 * direcciónes EEPROM usadas para guardar valores
 * 2, Estado de los semáforos 1 y 2
 * 3, Estado de los semáforos 3 y 4
 * 5,6, primera derección DCC
 * 
 * ------------------ SEMAFOROS y DIRECCIONES USADAS -------------------------------------------------------------------------------------------------------------------------
 * La máxima dirección posible es 1020 menos las direcciónes usadas 8, entonces la dirección inicial mas alta puede ser 1013 ya que se consumen 8 direcciones
 * El decoder consume 8 direcciónes consecitivas para controlar 4 semáforos de 4 aspectos (verde, amarillo, verderojo, amaillo-verde)
 * Dir + = semáforo 1 V
 * Dir - = semáforo 1 R
 * Dir+1 + = semáforo 1 RA
 * Dir+1 - = semáforo 1 A
 * Dir+2 + = semáforo 2 V
 * Dir+2 - = semáforo 2 R
 * Dir+3 + = semáforo 2 RA
 * Dir+3 - = semáforo 2 V
 * Dir+4 + = semáforo 3 V
 * Dir+4 - = semáforo 3 R
 * Dir+5 + = semáforo 3 RA
 * Dir+5 - = semáforo 3 A
 * Dir+6 + = semáforo 4 V
 * Dir+6 - = semáforo 4 R
 * Dir+7 + = semáforo 4 RA
 * Dir+7 - = semáforo 4 A
 * 
 * ------------------ LED's y SEMAFOROS ----------------------------------------------------------------------------------------------------------------------------------------
 * Si se conectan los LED diractamente, se ha de intercalar una resistencia en serie de entre 330ohm y 1K, uniendo los ánodos a +5V
 * Los semaforoa comerciales con LED tanbien funcionan con el ánodo común, pero a 16V, en este caso habrá que añadir la electrónica necesaría
 * 
 * ------------------ FUNCIONAMIENTO -------------------------------------------------------------------------------------------------------------------------------------------
 * INICIALIZACION:
 * La primera vez que ejecutamos el sketch en una placa ARDUINO, se encienden todos los LED de los semáforos y el LED de control esperando que enviemos una dirección DCC
 * esta se grabará como primera dirección del accesorio, si después de unos 15s no enviamos una dirección DCC por defecto inicializamos con dirección 1
 * esto pasará cada vez hasta que no grabemos una dirección DCC válida
 * 
 * VER DIRECCION DCC
 * Si accionammos el pulsador, se mostrará la dirección DCC grabada con destellos del LED de control, contando los destellos sabremos la dirección DCC
 * ejemplo, si la dirección es la 70, el LED de control parpadeará 7 veces, un intervalo apagado y un parpadeo largo para indicar 0
 * 
 * PROGRAMAR UNA NUEVA DIRECCION DCC:
 * Si mantenemos accionado el pulsador, después de unos 3s, el LED de control emitirá un destello, si soltamos entonces el pulsador se encenderá el LED de control 
 * esperando durante 15s, una nueva dirección para ser grabada
 *
 * PROGRAMAR TRANSICION ENTRE ESTADOS:
 * Si después del primer destello mantenemos el pulsador accionado, después de otros 3s, el LED de control emitirá otro destello, si soltamos el pulsador entonces se encenderá el LED de control  
 * Entonces podemos enviar una orden a una dirección DCC para definir el tiempo de transición entre estados, esto es la velocidad de encendido/apagado de los LED
 * El valor ha de estar entre 1 y 16, a menor valor mas lento
 * 
 * RESET DE VALORES:
 * Si al arrancar el decoder (o pulsar reset en el arduino) mantenemos activado el pulsador, se encenderá el LED de control durante unos 15s, si pulsamos de nuevo, se inicializarán 
 */
//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// inicialización común para la gestion DCC
#include <EEPROM.h>      // librería gestión EEPROM
const int PinDCC = 2;    // pin como entrada de señal DCC
const int NumAddrDCC = 8 ;  // numero de direcciónes consumidas
const int MaxDirDCC = 1020;  // dirección DCC mas alta admitida
int AddrBaseDCC;     //primera dirección DCC
byte contByte = 0;    // paquetes de bits recibidos
boolean datoLeido = false;     // control datos DCC leidos
boolean datoOK = false;   // datos DCC no disponibles
byte dato[4];                // valores leidos
byte longDato = 0;  // longitud datos leidos
byte contBit = 0;   // bits leidos
byte datoXor = 0;   // xor datos leidos 
unsigned int direccionLeida = 0;  // dirección DCC
unsigned int PdireccionLeida = 0; // dirección DCC previa
boolean PolDCC=false;  //  posicion  +/- , -= recto
boolean PPolDCC=false;  //  posicion previa
unsigned long DCCMicros = 0;  // tiempo para validar datos DCC 1/0
int DirDCC=0; // dirección de accesorio filtrada en rango de trabajo
boolean NoAccionDCC=false; //en proceso de capturar nueva dirección
//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// inicialización específica para el decoder de 4 semáforos de 4 estados
// usamos 6 bits para controlar el estado de cada par de sem´foros 1/2 y 3/4 los semáforos, 1=apagado, 0=encendido
byte Estado1 = EEPROM.read(2);  // Leemos último estado guardado para los semaforos 1 y 2
byte Estado2 = EEPROM.read(3);  // Leemos último estado guardado para los semaforos 3 y 4
byte P_Estado1 = 0;       // estado previo al cambio semaforos 1 y 2
byte P_Estado2 = 0;       // estado previo al cambio semaforos 3 y 4
#define PinLED A2   // pin para el LED de control
#define PinTECLA A0 // pin para tecla de control 
int LED [12] = {6,5,3 , 10,11,9 ,  4,12,15 , 17,18,19};  // Semaforo 1/2/3/4 , LED R-A-V/R-A-V/R-A-V/R-A-V
#define Periodo 1200
byte Incremento = EEPROM.read(4);  // aumentando este valor la transición entre estados se hace más rápida
int LEDnum;  // número de LED
int Contador;  // contador general
//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void setup() {   // inicializamos valores y realizamos verificaciones iniciales
  #if defined(DEBUG) 
    Serial.begin(9600);
    Serial.println("Decoder semaforos 4x3");
  #endif
  for (LEDnum = 0; LEDnum <12; LEDnum ++) {  // inicializamos pins de control del semáforo
    pinMode (LED [LEDnum], OUTPUT);
  } 
  pinMode(PinLED, OUTPUT);  // inicializamos pin del LED de control
  pinMode(PinTECLA, INPUT); // inicializamos tecla
  pinMode(A1, OUTPUT);  // por necesidades al diseñar el circuito impreso, usamos la salida A1 como referencia para al pull-up de la tecla, comentar para otros usos
  digitalWrite(A1, HIGH); // ponemos A1 a nivel alto (+5V)
  ReadAddDCC();
  if (AddrBaseDCC==0) SetAddDCC(); // si no hay dirección DCC definida, pedimos una
  if (AddrBaseDCC==0) { // si no hay dirección DCC definida definimos la #1 por defecto y inicializamos a verde los semáforos
    AddrBaseDCC = 1;  
    Estado1 = B011011; 
    Estado2 = B011011; 
    Incremento = 4;
  }  
  if (digitalRead(PinTECLA) == LOW) ResetValores();
  attachInterrupt(0, LeeDCC, CHANGE);  // interrupcion 0 en pin 2 ejecutamos dccdata() cada cambio de estado 
  SetEstado(0, Estado1, Estado1, P_Estado1, 2, 0);  // inicializamos con último estado guardado semaforos 1 y 2 
  SetEstado(0, Estado2, Estado2, P_Estado2, 3, 6);  // inicializamos con último estado guardado semaforos 3 y 4   
}

void loop() {
  if (datoOK == true) { // verificamos dato DCC correcto
    DirDCC=0;
    datoOK = false;
    detachInterrupt(0);  
    EvaluaDCC();         // evaluación de datos DCC leidos
    attachInterrupt(0, LeeDCC, CHANGE);
  } 
  if (digitalRead(PinTECLA) == LOW) { // controlamos si pulsamos 3s para mostrar dirección 
    Contador = 0;
    while (digitalRead(PinTECLA) == LOW ){ //&& Contador <10) { 
      //delay(1);
      Contador++;
      if (Contador==10) digitalWrite(PinLED, HIGH); delay(150); digitalWrite(PinLED, LOW); 
      if (Contador==20) digitalWrite(PinLED, HIGH); delay(150); digitalWrite(PinLED, LOW); 
    }     
    if (Contador<10) ShowAddDCC(); // pulsacion rapida <3s, mostramos dirección   
    if (Contador>10 & Contador<20) SetAddDCC(); // pulsacion media 3s a 6s, pedimos nueva dirección DCC
    if (Contador>=20 & Contador<30) SetIncremento(); // pulsación larga > 6s, pedimo nuevo valor para la transición entre LED's
  }  
}

void SetIncremento(){ // pasamos a la espera de una dirección DCC para configurar el decoder
  #if defined(DEBUG) 
    Serial.println("SetIncremento"); 
  #endif
  digitalWrite(PinLED, HIGH);  
  SetParmetro();
  if ((DirDCC > 0) && (DirDCC < 17)) {
    Incremento = DirDCC;
    EEPROM.write(4, Incremento);
  }
    #if defined(DEBUG) 
      Serial.print("Nuevo incremento: "); Serial.println(DirDCC);
    #endif
    digitalWrite(PinLED, LOW); delay(150); digitalWrite(PinLED, HIGH); delay(150); // anunciamos lectura valida con destellos
    digitalWrite(PinLED, LOW); delay(150); digitalWrite(PinLED, HIGH); delay(150); // anunciamos lectura valida con destellos
    digitalWrite(PinLED, LOW); delay(150); digitalWrite(PinLED, HIGH); delay(150); // anunciamos lectura valida con destellos
  digitalWrite(PinLED, LOW);  
}
 
void EjecutaAcciones (){ // ejecutamos segÃºn direcciÃ³n y polaridad leida
  // El 0 enciende, el 1 apaga, false = +, false = -
  // Gestion semaforos 1 y 2
  if ((DirDCC == AddrBaseDCC+0) && (PolDCC==false)) SetEstado(1, B011, Estado1, P_Estado1, 2, 0); //SEM1 Rxx
  if ((DirDCC == AddrBaseDCC+0) && (PolDCC==true))  SetEstado(1, B110, Estado1, P_Estado1, 2, 0); //SEM1 xxV
  if ((DirDCC == AddrBaseDCC+1) && (PolDCC==false)) SetEstado(1, B100, Estado1, P_Estado1, 2, 0); //SEM1 xAV
  if ((DirDCC == AddrBaseDCC+1) && (PolDCC==true))  SetEstado(1, B101, Estado1, P_Estado1, 2, 0); //SEM1 xAx 
  if ((DirDCC == AddrBaseDCC+2) && (PolDCC==false)) SetEstado(2, B011, Estado1, P_Estado1, 2, 0); //SEM2 Rxx
  if ((DirDCC == AddrBaseDCC+2) && (PolDCC==true))  SetEstado(2, B110, Estado1, P_Estado1, 2, 0); //SEM2 xxV
  if ((DirDCC == AddrBaseDCC+3) && (PolDCC==false)) SetEstado(2, B100, Estado1, P_Estado1, 2, 0); //SEM2 xAV
  if ((DirDCC == AddrBaseDCC+3) && (PolDCC==true))  SetEstado(2, B101, Estado1, P_Estado1, 2, 0); //SEM2 xAx
  // Gestion semaforos 3 y 4
  if ((DirDCC == AddrBaseDCC+4) && (PolDCC==false)) SetEstado(1, B011, Estado2, P_Estado2, 3, 6); //SEM1 Rxx
  if ((DirDCC == AddrBaseDCC+4) && (PolDCC==true))  SetEstado(1, B110, Estado2, P_Estado2, 3, 6); //SEM1 xxV
  if ((DirDCC == AddrBaseDCC+5) && (PolDCC==false)) SetEstado(1, B100, Estado2, P_Estado2, 3, 6); //SEM1 xAV
  if ((DirDCC == AddrBaseDCC+5) && (PolDCC==true))  SetEstado(1, B101, Estado2, P_Estado2, 3, 6); //SEM1 xAx 
  if ((DirDCC == AddrBaseDCC+6) && (PolDCC==false)) SetEstado(2, B011, Estado2, P_Estado2, 3, 6); //SEM2 Rxx
  if ((DirDCC == AddrBaseDCC+6) && (PolDCC==true))  SetEstado(2, B110, Estado2, P_Estado2, 3, 6); //SEM2 xxV
  if ((DirDCC == AddrBaseDCC+7) && (PolDCC==false)) SetEstado(2, B100, Estado2, P_Estado2, 3, 6); //SEM2 xAV
  if ((DirDCC == AddrBaseDCC+7) && (PolDCC==true))  SetEstado(2, B101, Estado2, P_Estado2, 3, 6); //SEM2 xAx
}
 
void SetEstado(int BlockSEM, byte EstadoSem, byte Estado, byte P_Estado, byte PosMem, byte SemOff){  // mostramos estado seleccionado
  switch(BlockSEM) { // filtramos segÃºn el semÃ¡foro que cambia
    case 0:  // cambio en los dos semaforos
      Estado = EstadoSem;
      break;
    case 1:  // cambio en semaforo 1/3
      Estado = (Estado & B000111) + (EstadoSem<<3);
      break;   
    case 2:  // cambio en semaforo 2/4
      Estado = (Estado & B111000) + EstadoSem;
      break;
  }
  if(Estado==P_Estado) return;                                // cancelamos si es el mismo estado que anterior
  for (LEDnum = SemOff; LEDnum < (6 + SemOff); LEDnum++) {   // Apagamos todos los LED's con estado = 1
    if(bitRead(Estado,LEDnum) != bitRead(P_Estado,LEDnum)) {  //sin cambio con respecto al estado anterior
      if (bitRead(P_Estado,LEDnum)==1) cambiaLED(LEDnum, HIGH);      // el estado anterior es encendido
    }
  }
  for (LEDnum = SemOff; LEDnum < (6 + SemOff); LEDnum++) {   // Encendemos todos los LED's con estado = 0
    if(bitRead(Estado,LEDnum) != bitRead(P_Estado,LEDnum)) {  //sin cambio con respecto al estado anterior
      if (bitRead(P_Estado,LEDnum)==0) cambiaLED(LEDnum, LOW);   // el estado anterior es apagado
    }
  }
  EEPROM.write(PosMem, Estado);    // Guardamos Estado en memoria 
  P_Estado1 = EEPROM.read(2); // Guardamos estado anterior semaforos 1 y 2
  P_Estado2 = EEPROM.read(3); // Guardamos estado anterior semaforos 3 y 4 
}

void cambiaLED(byte LEDnum, boolean Estado){
  Contador=0;
  while (Contador < Periodo){  // encendido o apagado lento
    Contador = Contador + Incremento;
    digitalWrite(LED[LEDnum], Estado);
    delayMicroseconds(Periodo - Contador);
    digitalWrite(LED[LEDnum], !Estado);
    delayMicroseconds(Contador);   
  } 
}

void ResetValores(){  // inicializamos valores
  digitalWrite(PinLED, HIGH);   
  while (digitalRead(PinTECLA) == LOW){}
  Contador = 0;
  while (Contador <30) { 
    delay(100);
    Contador++;
    if(digitalRead(PinTECLA) == LOW){  
      AddrBaseDCC = 1; // inicializamos dirección base a #1
      WriteAddDCC(); // guardamos nueva dirección DCC base
      EEPROM.write(2, B011011);
      EEPROM.write(3, B011011);
      EEPROM.write(4, 4);
      digitalWrite(PinLED, LOW);
      delay(2000);
      return;     
    } 
  }
  digitalWrite(PinLED, LOW); 
}
